package frc.robot.subsystems;

import java.util.function.BooleanSupplier;
import java.util.function.DoubleSupplier;

import com.revrobotics.CANSparkMax;
import com.revrobotics.RelativeEncoder;
import com.revrobotics.SparkPIDController;
import com.revrobotics.CANSparkBase.IdleMode;
import com.revrobotics.CANSparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj.DigitalInput;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ShooterConstants;

public class Shooter extends SubsystemBase{
    private CANSparkMax topShooter;
    private CANSparkMax bottomShooter;

    private RelativeEncoder bottomEncoder;
    private RelativeEncoder topEncoder;

    private CANSparkMax shooterPivot;

    private Encoder pivotEncoder;

    private SparkPIDController topPID;
    private SparkPIDController bottomPID;

    private PIDController shooterPivotPID;
    private ArmFeedforward shooterPivotFF;

    private DigitalInput shooterBeamBreak;

    private boolean indexerBeamBreak;

    public Shooter(BooleanSupplier indexerBeamBreak){
        topShooter = new CANSparkMax(ShooterConstants.kTopShooterID, MotorType.kBrushless);
        bottomShooter = new CANSparkMax(ShooterConstants.kBottomShooterID, MotorType.kBrushless);
        shooterPivot = new CANSparkMax(ShooterConstants.kShooterPivotID, MotorType.kBrushless);
        
        topEncoder = topShooter.getEncoder();
        bottomEncoder = bottomShooter.getEncoder();

        topPID = topShooter.getPIDController();
        topPID.setFeedbackDevice(topEncoder);
        bottomPID = bottomShooter.getPIDController();
        bottomPID.setFeedbackDevice(bottomEncoder);

        shooterPivot.setSmartCurrentLimit(50);
        topShooter.setSmartCurrentLimit(40);
        bottomShooter.setSmartCurrentLimit(40);

        pivotEncoder = new Encoder(0, 1);
        pivotEncoder.setDistancePerPulse(ShooterConstants.kPivotConversion);

        shooterBeamBreak = new DigitalInput(ShooterConstants.kShooterBeamBreakChannel);

        topShooter.setIdleMode(IdleMode.kBrake);
        bottomShooter.setIdleMode(IdleMode.kCoast);

        bottomShooter.burnFlash();
        shooterPivot.burnFlash();
        topShooter.burnFlash();

        topPID.setP(ShooterConstants.kShooterP);
        topPID.setI(ShooterConstants.kShooterI);
        topPID.setD(ShooterConstants.kShooterD);
        topPID.setFF(ShooterConstants.kShooterFF);

        bottomPID.setP(ShooterConstants.kShooterP);
        bottomPID.setI(ShooterConstants.kShooterI);
        bottomPID.setI(ShooterConstants.kShooterD);
        bottomPID.setFF(ShooterConstants.kShooterFF);

        shooterPivotPID = new PIDController(
            ShooterConstants.kShooterPivotP,
            ShooterConstants.kShooterPivotI,
            ShooterConstants.kShooterPivotD
        );

    }

    public Command angleSpeedsSetpoints(double setpointAngle, double topRPM, double bottomRPM){
        return run(()-> {
            angleAndSpeedControl(setpointAngle, topRPM, bottomRPM);
        });
    }

    private void angleAndSpeedControl(double setpointAngle, double topRPM, double bottomRPM){
        shooterPivot.setIdleMode(IdleMode.kBrake);

            shooterPivot.setVoltage(
                shooterPivotPID.calculate(pivotEncoder.getDistance(), setpointAngle) + 
                shooterPivotFF.calculate(setpointAngle, 0.0));

            topPID.setReference(topRPM, CANSparkMax.ControlType.kVelocity);
            bottomPID.setReference(bottomRPM, CANSparkMax.ControlType.kVelocity);
    }

    public Command ampHandoff(){
        return run(() -> {
            shooterPivot.setIdleMode(IdleMode.kBrake);

            shooterPivot.setVoltage(
                shooterPivotPID.calculate(pivotEncoder.getDistance(), ShooterConstants.kShooterLoadAngle) + 
                shooterPivotFF.calculate(ShooterConstants.kShooterLoadAngle, 0.0));
            
            if(shooterBeamBreak.get() || indexerBeamBreak){
                angleAndSpeedControl(ShooterConstants.kShooterLoadAngle, 1000.0, 1000.0);
            }else{
                angleAndSpeedControl(ShooterConstants.kShooterAmpAngle, 0, 0);
            }
        });
    }

    public Command climbState(){
        return run(() -> {
            shooterPivot.setIdleMode(IdleMode.kCoast);
            shooterPivot.set(0.0);
        });
    }

    public DoubleSupplier getShooterAngle(){
        return () -> {return pivotEncoder.getDistance();};
    }

    public BooleanSupplier getBeamBreak(){
        return shooterBeamBreak::get;
    }

    public Command manualPivot(DoubleSupplier pivotPower, DoubleSupplier spinny){
        return run(() ->{
            shooterPivot.set(pivotPower.getAsDouble());
            topShooter.set(spinny.getAsDouble());
            bottomShooter.set(spinny.getAsDouble());
        });
    }

}
